from distutils.core import setup

setup(
    name = "pyoddtool",
    version = "0.0.2",
    author = "Leonhard Lücken, Dmitry Puzyrev, Markus Kantner, Serhiy Yanchuk, Florian Stelzer",
    author_email = "uni.florian.stelzer@gmail.com",
    packages = ["pyoddtool"],
    package_data={'pyoddtool': ['oddtool/*', 'oddtool/sources/*']}
)